package com.fzm.common.service.impl;

import cn.fzm.chain.simplesdk.client.ParaChainClient;
import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.date.DateTime;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.util.IdUtil;
import cn.hutool.core.util.RandomUtil;
import cn.hutool.json.JSONUtil;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.UpdateWrapper;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.fzm.common.constant.SystemConstant;
import com.fzm.common.entity.AuthPerson;
import com.fzm.common.entity.User;
import com.fzm.common.entity.vo.UserListVo;
import com.fzm.common.entity.vo.UserStatisticVo;
import com.fzm.common.enums.AuthStatus;
import com.fzm.common.enums.AuthType;
import com.fzm.common.enums.IdCardVerificationResponse;
import com.fzm.common.enums.ResultCode;
import com.fzm.common.exception.GlobalException;
import com.fzm.common.interceptor.LoginUserInfo;
import com.fzm.common.mapper.UserMapper;
import com.fzm.common.mq.NotifyPublisher;
import com.fzm.common.params.LoginParam;
import com.fzm.common.properties.DebugProperties;
import com.fzm.common.properties.SmsProperties;
import com.fzm.common.service.AuthPersonService;
import com.fzm.common.service.UserService;
import com.fzm.common.utils.ObsUtil;
import com.fzm.common.utils.RedisUtil;
import com.fzm.common.utils.SmsUtil;
import com.fzm.common.utils.TencentApi;
import com.tencentcloudapi.faceid.v20180301.models.IdCardOCRVerificationResponse;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

import javax.servlet.http.HttpServletRequest;
import java.io.IOException;
import java.util.Base64;
import java.util.List;

/**
 * @author tangtuo
 * @date 2021/6/23 15:15
 */
@Slf4j
@Service
@Transactional(rollbackFor = Exception.class)
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class UserServiceImpl extends ServiceImpl<UserMapper, User> implements UserService {
    private final UserMapper userMapper;
    private final PasswordEncoder passwordEncoder;
    private final SmsProperties smsProperties;
    private final SmsUtil smsUtil;
    private final RedisUtil redisUtil;
    private final AuthPersonService authPersonService;
    private final ParaChainClient paraClient;
    private final HttpServletRequest request;
    private final ObsUtil obsUtil;
    private final DebugProperties debugProperties;
    private final NotifyPublisher notifyPublisher;


    @Override
    public User loadUserByUsername(String username) {
        QueryWrapper<User> wrapper = new QueryWrapper<>();
        wrapper.eq("telephone", username).
                or().
                eq("email", username);
        return getOne(wrapper);
    }


    @Override
    public User login(LoginParam param) {
        String username = StringUtils.isNotBlank(param.getTelephone()) ?
                param.getTelephone() :
                param.getEmail();
        User user = loadUserByUsername(username);

        if (user == null) {
            // 当前手机号或者邮箱没有注册过,走注册逻辑
            User registerUser = new User();
            BeanUtil.copyProperties(param, registerUser, true);
            // 生成钱包
            String wallet = paraClient.walletCreateAccount(IdUtil.simpleUUID());
            if (StringUtils.isBlank(wallet)) {
                throw GlobalException.newException(ResultCode.UNAUTHORIZED, "注册失败,生成区块链钱包异常,请稍后再试");
            }
            registerUser.setWallet(wallet);
            // 用户注册时生成默认昵称
            String nickname = "INMVO-" + RandomUtil.randomInt(100000, 999999);
            registerUser.setNickname(nickname);
            if (StringUtils.isNotBlank(param.getPassword())) {
                registerUser.setPassword(passwordEncoder.encode(param.getPassword()));
            } else {
                // 验证码注册流程
                if (!smsUtil.validateCode(getLoginCodeType(param.getCodetype()), username, param.getVerificationCode(), param.getCodetype())) {
                    throw GlobalException.newException(ResultCode.VALIDATE_FAILED, "验证码校验失败");
                }
            }
            save(registerUser);
            redisUtil.delete("user::statistic");
            return getById(registerUser.getId());
        }

        // 登录逻辑
        if (StringUtils.isNotBlank(param.getPassword())) {
            // 密码登录流程
            if (!passwordEncoder.matches(param.getPassword(), user.getPassword())) {
                throw GlobalException.newException(ResultCode.LOGIN_ERROR, "密码有误");
            }
        } else {
            // 验证码登录流程
            if (!smsUtil.validateCode(getLoginCodeType(param.getCodetype()), username, param.getVerificationCode(), param.getCodetype())) {
                throw GlobalException.newException(ResultCode.VALIDATE_FAILED, "验证码校验失败");
            }
        }

        String wallet = user.getWallet();
        if (StringUtils.isBlank(wallet)) {
            // 生成钱包
            wallet = paraClient.walletCreateAccount(IdUtil.simpleUUID());
            if (StringUtils.isBlank(wallet)) {
                throw GlobalException.newException(ResultCode.UNAUTHORIZED, "登录失败,生成区块链钱包异常,请稍后再试");
            }
            user.setWallet(wallet);

            User updateUser = new User();
            updateUser.setId(user.getId());
            updateUser.setWallet(wallet);
            updateById(updateUser);
        }

        // 认证成功的用户需要返回实名认证信息
        if (AuthStatus.SUCCESS.getStatus().equals(user.getAuthStatus())) {
            // 个人认证
            if (AuthType.PERSON.getType().equals(user.getAuthType())) {
                AuthPerson authPerson = authPersonService.getByUserId(user.getId());
                user.setAuthPerson(authPerson);
            }
            //TODO 企业认证
        }
        return user;
    }

    @Override
    public User getUserByToken() {
        String token = request.getHeader("Authorization");
        if (StringUtils.isBlank(token)) {
            throw GlobalException.newException(ResultCode.UNAUTHORIZED, "您还没有登录,请先登录");
        }
        Integer userId = LoginUserInfo.getLoginUser().getUserId();
        return getById(userId);
    }


    public String getLoginCodeType(String type) {
        switch (type) {
            case "sms":
                return smsProperties.getLoginMessageCodetype();
            case "email":
                return smsProperties.getLoginEmailCodetype();
            case "voice":
                return smsProperties.getLoginVoiceCodetype();
            default:
                throw GlobalException.newException(ResultCode.VALIDATE_FAILED, "短信类型【" + type + "】的传参有误,取值范围是{sms,email,voice}");
        }
    }

    @Override
    public User uploadAvatar(String avatar) {
        User loginUser = getUserByToken();
        // 如果不是第一次上传头像,需要删除旧头像
        if (StringUtils.isNotBlank(loginUser.getAvatar())) {
            obsUtil.delete(loginUser.getAvatar());
        }
        User user = new User().
                setAvatar(avatar).
                setId(loginUser.getId());
        updateById(user);
        return getById(loginUser.getId());
    }

    @Override
    @Cacheable(value = "user", key = "'statistic'")
    public UserStatisticVo getStatistic() {
        UserStatisticVo statisticVo = new UserStatisticVo();
        List<User> list = list();
        // 已注册用户数
        statisticVo.setRegisterUserCount(list.size());
        // 已发行NFT用户数
        long publishUserCount = list.stream().filter(user -> SystemConstant.BOOLEAN_DATA_TRUE.equals(user.getIsPublish())).count();
        statisticVo.setPublishUserCount((int) publishUserCount);
        // 已实名用户数
        long authUserCount = list.stream().filter(user -> AuthStatus.SUCCESS.getStatus().equals(user.getAuthStatus())).count();
        statisticVo.setAuthUserCount((int) authUserCount);
        // 个人实名用户数
        long personAuthCount = list.stream().
                filter(user -> AuthType.PERSON.getType().equals(user.getAuthType()) && AuthStatus.SUCCESS.getStatus().equals(user.getAuthStatus())).
                count();
        statisticVo.setPersonAuthCount((int) personAuthCount);
        // 企业实名用户数
        long enterpriseAuthCount = list.stream().
                filter(user -> AuthType.ENTERPRISE.getType().equals(user.getAuthType()) && AuthStatus.SUCCESS.getStatus().equals(user.getAuthStatus())).
                count();
        statisticVo.setEnterpriseAuthCount((int) enterpriseAuthCount);
        return statisticVo;
    }

    @Override
    public User getUserByWallet(String publishAddress) {
        QueryWrapper<User> queryWrapper = new QueryWrapper<>();
        queryWrapper.eq("wallet", publishAddress);
        return getOne(queryWrapper);
    }

    @Override
    public IdCardOCRVerificationResponse personAuth(MultipartFile cardPictureFront, MultipartFile cardPictureBack) throws IOException {
        User user = getUserByToken();
        if (AuthStatus.SUCCESS.getStatus().equals(user.getAuthStatus())) {
            throw GlobalException.newException(ResultCode.ID_CARD_VERIFICATION_ERROR, "您已认证成功,无法再次认证");
        }
        Integer status;
        Integer userId = user.getId();
        IdCardOCRVerificationResponse response;

        if (debugProperties.getAuth()) {
            response = new IdCardOCRVerificationResponse();
            response.setResult(IdCardVerificationResponse.SUCCESS.getCode());
            response.setIdCard("000000000000000000");
            response.setName("个人认证");
        } else {
            String encode = "data:image/jpg;base64," + Base64.getEncoder().encodeToString(cardPictureFront.getBytes());
            response = TencentApi.idCardOCRVerification(encode);
            log.info("实名认证结果:{}", JSONUtil.toJsonStr(response));
        }

        if (IdCardVerificationResponse.SUCCESS.getCode().equals(response.getResult())) {
            // 验证成功上传身份证证件
            String frontUrl = obsUtil.putObject(cardPictureFront);
            String backUrl = obsUtil.putObject(cardPictureBack);
            AuthPerson authPerson = new AuthPerson()
                    .setUserId(userId)
                    .setIdCard(response.getIdCard())
                    .setName(response.getName())
                    .setCardPictureBack(backUrl)
                    .setCardPictureFront(frontUrl);
            authPersonService.save(authPerson);
            status = AuthStatus.SUCCESS.getStatus();
        } else {
            status = AuthStatus.FAIL.getStatus();
        }
        UpdateWrapper<User> userUpdateWrapper = new UpdateWrapper<>();
        userUpdateWrapper.set("auth_type", AuthType.PERSON.getType())
                .set("auth_status", status)
                .eq("id", userId);
        update(userUpdateWrapper);
        if (AuthStatus.SUCCESS.getStatus().equals(status)) {
            redisUtil.delete("user::statistic");
            // 验证成功发送消息
            notifyPublisher.personAuthSuccessNotify(userId);
        } else {
            notifyPublisher.personAuthFailNotify(userId, response.getResult());
        }
        return response;
    }

    @Override
    public AuthPerson getAuthDetail(Integer userId) {
        User user = getById(userId);
        if (!AuthStatus.SUCCESS.getStatus().equals(user.getAuthStatus())) {
            throw GlobalException.newException(ResultCode.DATA_ERROR, "此用户实名认证未成功,无法查看其具体实名信息");
        }
        // 个人认证
        if (AuthType.PERSON.getType().equals(user.getAuthType())) {
            QueryWrapper<AuthPerson> queryWrapper = new QueryWrapper<>();
            queryWrapper.eq("user_id", userId);
            return authPersonService.getOne(queryWrapper);
        }
        return null;
    }

    @Override
    public List<UserListVo> list(String telephone, String nickname, Integer authType, String start, String end) {
        DateTime startDate = null;
        DateTime endDate = null;
        if (StringUtils.isNotBlank(start)) {
            startDate = DateUtil.parse(start + " 00:00:00", "yyyy-MM-dd HH:mm:ss");
        }
        if (StringUtils.isNotBlank(end)) {
            endDate = DateUtil.parse(end + " 23:59:59", "yyyy-MM-dd HH:mm:ss");
        }
        return userMapper.list(telephone, nickname, authType, startDate, endDate);
    }

}
